import argparse
import json
import os
import sys
from typing import cast

from badabump import __app__, __version__
from badabump.annotations import Argv
from badabump.cleaners import clean_body, clean_commit_subject, clean_tag_ref
from badabump.cli.arguments import add_path_argument
from badabump.cli.output import github_actions_output
from badabump.configs import ProjectConfig
from badabump.git import Git
from badabump.regexps import to_regexp
from badabump.versions import Version


def parse_args(argv: Argv) -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        prog=f"{__app__}-ci",
        description=(
            "Useful commands for dealing with release commits & tags at CI"
        ),
    )
    parser.add_argument(
        "-v", "--version", action="version", version=__version__
    )
    add_path_argument(parser)

    subparsers = parser.add_subparsers()

    prepare_tag_parser = subparsers.add_parser("prepare_tag")
    prepare_tag_parser.set_defaults(func=prepare_tag)

    prepare_release_parser = subparsers.add_parser("prepare_release")
    prepare_release_parser.add_argument(
        "ref",
        default=os.getenv("GITHUB_REF"),
        help="Tag reference. By default: GITHUB_REF env var",
        metavar="REF",
        nargs="?",
    )
    prepare_release_parser.set_defaults(func=prepare_release)

    return parser.parse_args(argv)


def prepare_release(args: argparse.Namespace, *, config: ProjectConfig) -> int:
    git = Git(path=config.path)

    tag_ref = clean_tag_ref(args.ref)
    version = Version.from_tag(tag_ref, config=config)
    github_actions_output("tag_name", tag_ref)
    github_actions_output(
        "is_pre_release", json.dumps(version.pre_release is not None)
    )

    tag_subject = git.retrieve_tag_subject(tag_ref)
    github_actions_output("release_name", tag_subject)

    tag_body = git.retrieve_tag_body(tag_ref)
    github_actions_output("release_body", tag_body)

    return 0


def prepare_tag(args: argparse.Namespace, *, config: ProjectConfig) -> int:
    git = Git(path=config.path)
    git_commit = git.retrieve_last_commit()
    try:
        raw_subject, _, *body = git_commit.splitlines()
    except ValueError:
        print("ERROR: Last commit has empty body. Exit...", file=sys.stderr)
        return 1

    expected_re = to_regexp(config.pr_title_format)
    matched = expected_re.match(clean_commit_subject(raw_subject))
    if matched is None:
        print(
            "ERROR: Last commit has unexpected subject line. Exit...",
            file=sys.stderr,
        )
        return 1

    version = matched.groupdict()["version"]
    github_actions_output(
        "tag_name", config.tag_format.format(version=version)
    )
    github_actions_output(
        "tag_message",
        "\n\n".join(
            (
                config.tag_subject_format.format(version=version),
                clean_body(body),
            )
        ),
    )

    return 0


def main(argv: Argv = None) -> int:
    args = parse_args(argv or sys.argv[1:])
    # TODO: Fix this by providing required flag on adding subparsers
    if getattr(args, "func", None) is None:
        print(
            "ERROR: Please provide one of available subcommands. Exit...",
            file=sys.stderr,
        )
        return 1

    config = ProjectConfig.from_path(args.path)
    return cast(int, args.func(args, config=config))



import json
from contextlib import suppress
from typing import cast, Optional, Type, Union

import attr
import tomli

from badabump.annotations import DictStrStr
from badabump.configs import ProjectConfig, UpdateConfig
from badabump.enums import ProjectTypeEnum, VersionTypeEnum
from badabump.regexps import to_regexp
from badabump.versions import calver, pre_release, semver
from badabump.versions.calver import CalVer
from badabump.versions.exceptions import VersionError, VersionParseError
from badabump.versions.parsing import parse_version
from badabump.versions.pre_release import PreRelease
from badabump.versions.semver import SemVer


CalOrSemVer = Union[CalVer, SemVer]


@attr.dataclass(frozen=True, slots=True)
class Version:
    version: CalOrSemVer
    pre_release: Optional[PreRelease] = None

    @classmethod
    def from_tag(cls, value: str, *, config: ProjectConfig) -> "Version":
        return cls.parse(
            guess_version_from_tag(value, tag_format=config.tag_format),
            config=config,
        )

    def format(self, *, config: ProjectConfig) -> str:  # noqa: A003
        if self.pre_release:
            return (
                f"{self.version.format()}"
                f"{self.pre_release.format(project_type=config.project_type)}"
            )
        return self.version.format()

    @classmethod
    def guess_initial_version(
        cls, *, config: ProjectConfig, is_pre_release: bool
    ) -> "Version":
        version = guess_initial_version(config)
        if version.pre_release is None and is_pre_release:
            return attr.evolve(version, pre_release=PreRelease())
        return version

    @classmethod
    def parse(cls, value: str, *, config: ProjectConfig) -> "Version":
        schema = config.version_schema

        version_cls: Type[CalOrSemVer]
        full_schema_parts: DictStrStr
        if config.version_type == VersionTypeEnum.semver:
            version_cls = semver.SemVer
            full_schema_parts = semver.SCHEMA_PARTS_PARSING
        else:
            version_cls = calver.CalVer
            full_schema_parts = calver.SCHEMA_PARTS_PARSING

        with suppress(VersionError):
            return cls(version=version_cls.parse(value, schema=schema))

        full_schema = (
            f"{schema}{pre_release.SCHEMA_MAPPING[config.project_type]}"
        )
        full_schema_parts.update(pre_release.SCHEMA_PARTS_PARSING)

        maybe_parsed = parse_version(full_schema, full_schema_parts, value)
        if maybe_parsed:
            return cls(
                version=version_cls.from_parsed_dict(
                    maybe_parsed, schema=schema
                ),
                pre_release=pre_release.PreRelease.from_parsed_dict(
                    maybe_parsed, project_type=config.project_type
                ),
            )

        raise VersionParseError(schema, value)

    def update(self, config: UpdateConfig) -> "Version":
        if self.pre_release:
            return Version(
                version=self.version,
                pre_release=self.pre_release.update(config),
            )

        if config.is_pre_release:
            return Version(
                version=self.version.update(
                    attr.evolve(config, is_pre_release=False)
                ),
                pre_release=PreRelease(),
            )

        return Version(version=self.version.update(config))


def find_project_version(config: ProjectConfig) -> Optional[str]:
    if config.project_type == ProjectTypeEnum.javascript:
        package_json_path = config.path / "package.json"
        if package_json_path.exists():
            try:
                return cast(
                    str, json.loads(package_json_path.read_text())["version"]
                )
            except (KeyError, ValueError):
                ...
    else:
        pyproject_toml_path = config.path / "pyproject.toml"
        if pyproject_toml_path.exists():
            try:
                return cast(
                    str,
                    tomli.loads(pyproject_toml_path.read_text())["tool"][
                        "poetry"
                    ]["version"],
                )
            except (KeyError, ValueError):
                ...

    return None


def guess_initial_version(config: ProjectConfig) -> Version:
    maybe_version_str = find_project_version(config)
    if maybe_version_str:
        return Version.parse(maybe_version_str, config=config)

    if config.version_type == VersionTypeEnum.semver:
        return Version(version=SemVer.initial())
    return Version(version=CalVer.initial(schema=config.version_schema))


def guess_version_from_tag(value: str, *, tag_format: str) -> str:
    matched = to_regexp(tag_format).match(value)
    if matched:
        return matched.groupdict()["version"]
    raise ValueError("Tag value does not match given tag format")



from contextlib import suppress
from enum import Enum, unique
from typing import DefaultDict, Optional

import attr

from badabump.annotations import DictStrAny, DictStrStr
from badabump.configs import UpdateConfig
from badabump.enums import ProjectTypeEnum
from badabump.versions.formatting import format_version
from badabump.versions.parsing import parse_version


SCHEMA_MAPPING: DefaultDict[ProjectTypeEnum, str] = DefaultDict(
    lambda: "-TYPE.NUMBER"
)
SCHEMA_MAPPING[ProjectTypeEnum.python] = "TYPENUMBER"

SCHEMA_PARTS_FORMATTING = {
    "TYPE": "{type}",
    "NUMBER": "{number}",
}
SCHEMA_PARTS_PARSING = {
    "TYPE": r"(?P<type>\D+)",
    "NUMBER": r"(?P<number>\d+)",
}


@unique
class PreReleaseTypeEnum(Enum):
    alpha = "alpha"
    beta = "beta"
    rc = "rc"


NEXT_PRE_RELEASE_TYPE = {
    PreReleaseTypeEnum.alpha: PreReleaseTypeEnum.beta,
    PreReleaseTypeEnum.beta: PreReleaseTypeEnum.rc,
    PreReleaseTypeEnum.rc: PreReleaseTypeEnum.rc,
}

PRE_RELEASE_TYPE_MAPPING = {
    ProjectTypeEnum.python: {
        PreReleaseTypeEnum.alpha: "a",
        PreReleaseTypeEnum.beta: "b",
        PreReleaseTypeEnum.rc: "rc",
    },
}


@attr.dataclass(frozen=True, slots=True)
class PreRelease:
    pre_release_type: PreReleaseTypeEnum = PreReleaseTypeEnum.alpha
    number: int = 0

    def format(self, *, project_type: ProjectTypeEnum) -> str:  # noqa: A003
        schema = SCHEMA_MAPPING[project_type]

        context: DictStrAny = {"number": self.number}
        maybe_type_mapping = PRE_RELEASE_TYPE_MAPPING.get(project_type)
        if maybe_type_mapping:
            context["type"] = maybe_type_mapping[self.pre_release_type]
        else:
            context["type"] = self.pre_release_type.value

        return format_version(schema, SCHEMA_PARTS_FORMATTING, context)

    @classmethod
    def from_parsed_dict(
        cls, parsed_dict: DictStrStr, *, project_type: ProjectTypeEnum
    ) -> "PreRelease":
        return cls(
            pre_release_type=guess_pre_release_type(
                parsed_dict["type"], project_type=project_type
            ),
            number=int(parsed_dict["number"]),
        )

    @classmethod
    def parse(
        cls, value: str, *, project_type: ProjectTypeEnum
    ) -> "PreRelease":
        schema = SCHEMA_MAPPING[project_type]

        maybe_parsed = parse_version(schema, SCHEMA_PARTS_PARSING, value)
        if maybe_parsed:
            with suppress(KeyError):
                return cls.from_parsed_dict(
                    maybe_parsed, project_type=project_type
                )

        raise ValueError(
            "Invalid pre-release value, which do not match any supported "
            "project type"
        )

    def update(self, config: UpdateConfig) -> Optional["PreRelease"]:
        if config.is_pre_release is False:
            return None

        if config.is_breaking_change:
            next_type = NEXT_PRE_RELEASE_TYPE[self.pre_release_type]
            if next_type != self.pre_release_type:
                return PreRelease(pre_release_type=next_type, number=0)
            return PreRelease(
                pre_release_type=next_type, number=self.number + 1
            )

        return PreRelease(
            pre_release_type=self.pre_release_type, number=self.number + 1
        )


def guess_pre_release_type(
    value: str, *, project_type: ProjectTypeEnum
) -> PreReleaseTypeEnum:
    maybe_type_mapping = PRE_RELEASE_TYPE_MAPPING.get(project_type)
    if maybe_type_mapping:
        return {
            item: pre_release_type
            for pre_release_type, item in maybe_type_mapping.items()
        }[value]
    return PreReleaseTypeEnum[value]



import attr

from badabump.annotations import DictStrStr
from badabump.configs import UpdateConfig
from badabump.constants import DEFAULT_SEMVER_SCHEMA as SCHEMA
from badabump.versions.exceptions import VersionParseError
from badabump.versions.formatting import format_version
from badabump.versions.parsing import parse_version


SCHEMA_PARTS_FORMATTING = {
    "MAJOR": "{major}",
    "MINOR": "{minor}",
    "PATCH": "{patch}",
}
SCHEMA_PARTS_PARSING = {
    "MAJOR": r"(?P<major>\d+)",
    "MINOR": r"(?P<minor>\d+)",
    "PATCH": r"(?P<patch>\d+)",
}


@attr.dataclass(frozen=True, slots=True)
class SemVer:
    major: int
    minor: int
    patch: int

    schema: str = SCHEMA

    def format(self) -> str:  # noqa: A003
        return format_version(
            self.schema, SCHEMA_PARTS_FORMATTING, attr.asdict(self)
        )

    @classmethod
    def from_parsed_dict(
        cls, parsed: DictStrStr, *, schema: str = None
    ) -> "SemVer":
        return cls(
            major=int(parsed["major"]),
            minor=int(parsed["minor"]),
            patch=int(parsed["patch"]),
            schema=schema or SCHEMA,
        )

    @classmethod
    def initial(cls, *, schema: str = None) -> "SemVer":
        return cls(major=1, minor=0, patch=0, schema=schema or SCHEMA)

    @classmethod
    def parse(cls, value: str, *, schema: str = None) -> "SemVer":
        maybe_parsed = parse_version(SCHEMA, SCHEMA_PARTS_PARSING, value)
        if maybe_parsed:
            return cls.from_parsed_dict(maybe_parsed, schema=schema)
        raise VersionParseError(schema or SCHEMA, value)

    def update(self, config: UpdateConfig) -> "SemVer":
        if config.is_pre_release:
            return self

        if config.is_breaking_change:
            return attr.evolve(self, major=self.major + 1, minor=0, patch=0)

        if config.is_minor_change:
            return attr.evolve(
                self, major=self.major, minor=self.minor + 1, patch=0
            )

        return attr.evolve(self, patch=self.patch + 1)



import subprocess
from contextlib import suppress
from pathlib import Path
from typing import Iterator, List, Optional, Tuple

import attr


@attr.dataclass(frozen=True, slots=True)
class Git:
    path: Path

    def list_commits(self, from_ref: str) -> Tuple[str, ...]:
        def iter_commtis(commit_ids: List[str]) -> Iterator[str]:
            for commit_id in commit_ids:
                yield self._check_output(
                    ["git", "log", "-1", "--format=%B", commit_id]
                )

        commit_ids = self._check_output(
            ["git", "log", "--format=%H", f"{from_ref}..HEAD"]
        ).splitlines()
        return tuple(iter_commtis(commit_ids))

    def retrieve_last_commit(self) -> str:
        return self._check_output(["git", "log", "-1", "--format=%B"])

    def retrieve_last_tag(self) -> str:
        return self._check_output(["git", "describe", "--abbrev=0", "--tags"])

    def retrieve_last_tag_or_none(self) -> Optional[str]:
        with suppress(subprocess.CalledProcessError, ValueError):
            return self.retrieve_last_tag()
        return None

    def retrieve_tag_body(self, tag: str) -> str:
        return self._check_output(
            ["git", "tag", "-l", "--format=%(body)", tag]
        )

    def retrieve_tag_subject(self, tag: str) -> str:
        return self._check_output(
            ["git", "tag", "-l", "--format=%(subject)", tag]
        )

    def _check_output(self, args: List[str]) -> str:
        maybe_output = subprocess.check_output(args, cwd=self.path)
        if maybe_output is not None:
            return maybe_output.strip().decode("utf-8")

        raise ValueError("git command return unexpected empty output")



from badabump import __app__
from badabump.enums import FormatTypeEnum, ProjectTypeEnum, VersionTypeEnum


CHANGELOG_UPPER = "CHANGELOG"
CHANGELOG_LOWER = CHANGELOG_UPPER.lower()

DEFAULT_CHANGELOG_FILE_INCLUDE_DATE = True
DEFAULT_CHANGELOG_FORMAT_TYPE_FILE = FormatTypeEnum.markdown
DEFAULT_CHANGELOG_FORMAT_TYPE_GIT = FormatTypeEnum.markdown
DEFAULT_CHANGELOG_IGNORE_FOOTER_URLS = True
DEFAULT_PROJECT_TYPE = ProjectTypeEnum.python
DEFAULT_TAG_FORMAT = "v{version}"
DEFAULT_TAG_SUBJECT_FORMAT = "{version} Release"
DEFAULT_PR_BRANCH_FORMAT = f"chore/release-{DEFAULT_TAG_FORMAT}"
DEFAULT_PR_TITLE_FORMAT = f"chore: {DEFAULT_TAG_SUBJECT_FORMAT}"
DEFAULT_VERSION_TYPE = VersionTypeEnum.calver
DEFAULT_STRICT_MODE = False

DEFAULT_CALVER_SCHEMA = DEFAULT_VERSION_SCHEMA = "YY.MINOR.MICRO"
DEFAULT_SEMVER_SCHEMA = "MAJOR.MINOR.PATCH"

INITIAL_RELEASE_COMMIT = "feat: Initial release"
INITIAL_PRE_RELEASE_COMMIT = "feat: Initial pre-release"

FILE_CONFIG_TOML = f".{__app__}.toml"
FILE_PACKAGE_JSON = "package.json"
FILE_PACKAGE_LOCK_JSON = "package-lock.json"
FILE_PYPROJECT_TOML = "pyproject.toml"
FILE_YARN_LOCK = "yarn.lock"



from badabump.versions.calver import CalVer
from badabump.versions.pre_release import PreRelease
from badabump.versions.semver import SemVer
from badabump.versions.version import Version


__all__ = ("CalVer", "PreRelease", "SemVer", "Version")



from difflib import ndiff


EMPTY = "-"
VALUE_ESCAPE_MAPPING = (("%", "%25"), ("\n", "%0A"), ("\r", "%0D"))


def diff(current_content: str, next_content: str) -> str:
    items = ndiff(
        current_content.splitlines(keepends=True),
        next_content.splitlines(keepends=True),
    )
    return "".join(item for item in items if item.startswith(("-", "+", "?")))


def echo_message(message: str, *, is_dry_run: bool) -> None:
    prefix = "[DRY-RUN] " if is_dry_run else ""
    print(f"{prefix}{message}")


def echo_value(
    label: str, value: str, *, is_ci: bool = False, ci_name: str = None
) -> None:
    if is_ci and ci_name:
        github_actions_output(ci_name, value)
    else:
        print(f"{label}{value}")


def github_actions_output(name: str, value: str) -> None:
    for symbol, code in VALUE_ESCAPE_MAPPING:
        value = value.replace(symbol, code)
    print(f"::set-output name={name}::{value}")
